#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Light.h"
#include "ngl/Matrix.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/Material.h"
#include "ngl/NGLInit.h"
#include "ngl/VBOPrimitives.h"
#include "ngl/ShaderManager.h"
#include "ngl/Random.h"
#include <boost/foreach.hpp>
#include "ngl/Util.h"


//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
									 int _numSpheres,
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

  // set this widget to have the initial keyboard focus
  setFocus();
	m_numSpheres=_numSpheres;
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // Now set the initial GLWindow attributes to default values
  // Roate is false
  m_rotate=false;
  // mouse rotation values set to 0
  m_spinXFace=0;
  m_spinYFace=0;
  m_origX=0;
  m_origY=0;
  m_rayUpdateTimer=startTimer(10);
  m_animate=true;
	// now create the actual spheres for our program
	ngl::Random *rng=ngl::Random::instance();
	float x;
	float y;

	for (int i=0; i<m_numSpheres; ++i)
	{
		x=rng->randomNumber(10);
		y=rng->randomNumber(8);
		m_sphereArray.push_back(Sphere(ngl::Vector(x,y,0),rng->randomPositiveNumber(1)+0.2));
	}

	// create the points for our ray
	m_rayStart.set(0,10,0);
	m_rayEnd.set(0,-5,0);
	m_rayStart2.set(0,0,20);
	m_rayEnd2.set(0,0,-5);


}

GLWindow::~GLWindow()
{
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  std::cout<<"Shutting down NGL, removing VBO's and Shaders\n";
  Init->NGLQuit();
}

// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{
  // create the lights

  glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
  // enable depth testing for drawing
   glEnable(GL_DEPTH_TEST);
   // Now we will create a basic Camera from the graphics library
   // This is a static camera so it only needs to be set once
   // First create Values for the camera position
   ngl::Vector From(0,0,-25);
   ngl::Vector To(0,0,0);
   ngl::Vector Up(0,1,0);
   ngl::NGLInit *Init = ngl::NGLInit::instance();
   #ifdef WIN32
  	glewInit();
  #endif
  Init->initGlew();
  m_cam= new ngl::Camera(From,To,Up,ngl::PERSPECTIVE);
   // set the shape using FOV 45 Aspect Ratio based on Width and Height
   // The final two are near and far clipping planes of 0.5 and 10
   m_cam->setShape(45,(float)720.0/576.0,0.5,150,ngl::PERSPECTIVE);
   // now to load the shader and set the values
   // grab an instance of shader manager
   ngl::ShaderManager *shader=ngl::ShaderManager::instance();
   // load a frag and vert shaders
   shader->loadShader("gl3xTest","shaders/Vertex.vs","shaders/Fragment.fs");
   // set this as the active shader
   shader->useShader("gl3xTest");
   // now pass the modelView and projection values to the shader
   shader->setShaderParamFromMatrix("gl3xTest","ViewMatrix",m_cam->getModelView());
   shader->setShaderParamFromMatrix("gl3xTest","projectionMatrix",m_cam->getProjection());
   shader->setShaderParam1i("gl3xTest","Normalize",true);
   shader->setShaderParam1i("gl3xTest","numLightsEnabled",8);
   shader->setShaderParam4f("gl3xTest","eye",10,4,2,1);

   glEnable(GL_DEPTH_TEST); // for removal of hidden surfaces

   ngl::VBOPrimitives *prim =  ngl::VBOPrimitives::instance();
   prim->createVBOSphere("sphere",1.0,40);
	 prim->createVBOSphere("smallSphere",0.2,10);


}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  glViewport(0,0,_w,_h);
  m_cam->setShape(45,(float)_w/_h,0.5,150,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  shader->setShaderParamFromMatrix("gl3xTest","projectionMatrix",m_cam->getProjection());
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 // build our transform stack
  ngl::Transformation trans;
  // set the mouse rotation
  trans.setRotation(m_spinXFace,m_spinYFace,0);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);

  // grab an instance of the primitives for drawing
  ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();
  // Rotation based on the mouse position for our global
  // transform
  // grab an instance of the shader manager
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	shader->use("gl3xTest");
	ngl::Material m(ngl::GOLD);
	m.use();

	m_transformStack.loadGlobalAndCurrentMatrixToShader("gl3xTest","ModelMatrix");
  shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());

	// draw a cube at the ray start points
	m_transformStack.pushTransform();
	{
		m_transformStack.getCurrentTransform().setPosition(m_rayStart);

		shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
		prim->draw("cube");
	} // and before a pop
	m_transformStack.popTransform();

	m_transformStack.pushTransform();
	{
		m_transformStack.getCurrentTransform().setPosition(m_rayStart2);

		shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
		prim->draw("cube");
	} // and before a pop
	m_transformStack.popTransform();




	BOOST_FOREACH(Sphere s, m_sphereArray)
	{
		m.use();
		s.draw("gl3xTest",m_transformStack);
		if(s.isHit())
		{
			ngl::Vector dir=m_rayEnd-m_rayStart;
			ngl::Vector dir2=m_rayEnd2-m_rayStart2;
		  drawHitPoints(m_rayStart,dir,s.getPos(),s.getRadius());
		  drawHitPoints(m_rayStart2,dir2,s.getPos(),s.getRadius());
		}
	}
	// we build up a vertex array for the lines of the start and end points and draw
	m_transformStack.pushTransform();
	{
		shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());

		glEnableClientState(GL_VERTEX_ARRAY);
		GLfloat points[3*4];
		points[0]=m_rayStart.openGL()[0];
		points[1]=m_rayStart.openGL()[1];
		points[2]=m_rayStart.openGL()[2];
		points[3]=m_rayEnd.openGL()[0];
		points[4]=m_rayEnd.openGL()[1];
		points[5]=m_rayEnd.openGL()[2];

		points[6]=m_rayStart2.openGL()[0];
		points[7]=m_rayStart2.openGL()[1];
		points[8]=m_rayStart2.openGL()[2];
		points[9]=m_rayEnd2.openGL()[0];
		points[10]=m_rayEnd2.openGL()[1];
		points[11]=m_rayEnd2.openGL()[2];
		glVertexPointer(3, GL_FLOAT, 0, points);
		glDrawArraysInstancedARB(GL_LINES, 0, 4,1);
		glDisableClientState(GL_VERTEX_ARRAY);
	}
	m_transformStack.popTransform();


}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::drawHitPoints(
															ngl::Vector _raystart,
															ngl::Vector _raydir,
															ngl::Vector _pos,
															GLfloat _radius
														)
{
GLfloat A,B,C,discrim;
ngl::Vector p;
GLfloat discRoot;
GLfloat t1 ;        // earlier hit
GLfloat t2 ; 		// second hit
// points for the hit points
ngl::Vector h1 ;
ngl::Vector h2 ;

// normalize the ray
_raydir.normalize();
// cal the A value as the dotproduct a.a (see lecture notes)
A = _raydir.dot(_raydir);
//b= 2*d.(Po-Pc)
p=_raystart-_pos;
B= _raydir.dot(p)*2.0;
// C = (Po-Pc).(Po-Pc)-r^2
C=p.dot(p)-_radius*_radius;
// finally get the descrim
// b^2-4(ac)
discrim = B * B - 4.0*(A * C);
// if the discrim <= 0.0 it's not a hit
if(discrim >= 0.0)
{
	discRoot = sqrt(discrim);
	t1 = (-B-(discRoot))/2*A;        // earlier hit
	t2 = (-B+(discRoot))/2*A; 		// second hit
	// move along the ray from origin in ray direction with an offset of t1 / t2
	// to get the hit points
  h1=_raystart + (_raydir*t1);
  h2=_raystart + (_raydir*t2);
	// draw the hit points
	ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();
	ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	ngl::Material m(ngl::BLACKPLASTIC);
  m.use();
	m_transformStack.pushTransform();
	{
		m_transformStack.getCurrentTransform().setPosition(h1);
		shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
		prim->draw("smallSphere");
	} // and before a pop
	m_transformStack.popTransform();

	m_transformStack.pushTransform();
	{
		m_transformStack.getCurrentTransform().setPosition(h2);

		shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
		prim->draw("smallSphere");
	} // and before a pop
	m_transformStack.popTransform();
}
}
//----------------------------------------------------------------------------------------------------------------------
bool GLWindow::raySphere(
													ngl::Vector _rayStart,
													ngl::Vector _rayDir,
													ngl::Vector _pos,
													GLfloat _radius
												)
{
// variables for the Quadratic roots and discriminator
GLfloat A,B,C,discrim;
ngl::Vector p;
// normalize the ray
_rayDir.normalize();
// cal the A value as the dotproduct a.a (see lecture notes)
A = _rayDir.dot(_rayDir);
//b= 2*d.(Po-Pc)
p=_rayStart-_pos;
B= _rayDir.dot(p)*2;
// C = (Po-Pc).(Po-Pc)-r^2
C=p.dot(p)-_radius*_radius;
// finally get the descrim
// b^2-4(ac)
discrim = B * B - 4*(A * C);
// if the discrim <= 0.0 it's not a hit
if(discrim <= 0.0)
{
	return false;
}
else
{
	return true;
}
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::update()
{
enum {FWD,BWD};

static int s_direction=0;
bool collide;
ngl::Vector dir;
ngl::Vector dir2;
dir=m_rayEnd-m_rayStart;
dir2=m_rayEnd2-m_rayStart2;

// note here we need to iterate by reference as we want to modify the Sphere Objects
// so we need to explicitly create our object as a reference object
BOOST_FOREACH(Sphere &s,m_sphereArray)
{
	s.setNotHit();
	collide =raySphere(m_rayStart,dir,s.getPos(),s.getRadius());
	if(collide)
	{
		s.setHit();
	}
	collide =raySphere(m_rayStart2,dir2,s.getPos(),s.getRadius());
	if(collide)
	{
		s.setHit();
	}
}

// now update the rays
if(s_direction==FWD)
{
	m_rayEnd.m_x+=0.1;
	m_rayEnd2.m_x-=0.1;
	if(m_rayEnd.m_x>22.0)
	{
		s_direction=BWD;
	}
}
else
{
	m_rayEnd.m_x-=0.1;
	m_rayEnd2.m_x+=0.1;
	if(m_rayEnd.m_x<=-22.0)
	{
		s_direction=FWD;
	}
 }
}



//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}

void GLWindow::timerEvent(
                          QTimerEvent *_event
                         )
{
	if(_event->timerId() == m_rayUpdateTimer)
	{
		if (m_animate !=true)
		{
			return;
		}
		update();
		updateGL();
	}
}
