#ifndef __GL_WINDOW_H__
#define __GL_WINDOW_H__

#include "Triangle.h"
#include "ngl/Camera.h"
#include "ngl/Colour.h"
#include "ngl/TransformStack.h"
#include "ngl/Vector.h"
// must be included after our stuff becuase GLEW needs to be first
#include <QtOpenGL>


class GLWindow : public QGLWidget
{
Q_OBJECT        // must include this if you use Qt signals/slots
public :
  /// @brief Constructor for GLWindow
  /// @param [in] parent the parent window to create the GL context in
  GLWindow(
					 int _numTriangles,
           QWidget *_parent
          );
  ~GLWindow();
	//----------------------------------------------------------------------------------------------------------------------
	/// @brief override the keyPressEvent inherited from QObject so we can handle key presses.
  /// @param [in] _event the event to process

	void processKey(
                     QKeyEvent *_event
                  );
private :

  /// @brief used to store the x rotation mouse value
  int m_spinXFace;
  /// @brief used to store the y rotation mouse value
  int m_spinYFace;
  /// @brief flag to indicate if the mouse button is pressed when dragging
  bool m_rotate;
  /// @brief the previous x mouse value
  int m_origX;
  /// @brief the previous y mouse value
  int m_origY;
  /// @brief Our Camera
  ngl::Camera *m_cam;
   /// @brief our transform stack for drawing the teapots
  ngl::TransformStack m_transformStack;

	/// @brief our spheres to test against
	std::vector<Triangle> m_triangleArray;
	/// @brief number of spheres
	int m_numTriangles;

	ngl::Vector m_rayStart;
	ngl::Vector m_rayEnd;



protected:

  /// \brief  The following methods must be implimented in the sub class
  /// this is called when the window is created
  /// \note these are part of the Qt API so can't be changed to the coding standard
  /// so it can't be called initializeGL )
  void initializeGL();

  /// \brief this is called whenever the window is re-sized
  /// @param[in] _w the width of the resized window
  /// @param[in] _h the height of the resized window
  /// \note these are part of the Qt API so can't be changed to the coding standard
  /// so it can't be called resizeGL )
  void resizeGL(
                const int _w,
                const int _h
               );
  /// @brief this is the main gl drawing routine which is called whenever the window needs to
  /// be re-drawn
  void paintGL();

private :



  /// @brief this method is called every time a mouse is moved
  /// @param _event the Qt Event structure
  /// @note these are part of the Qt API so can't be changed to the coding standard
  /// so it can't be called MouseMoveEvent )

  void mouseMoveEvent (
                       QMouseEvent * _event
                      );
  /// \brief this method is called everytime the mouse button is pressed
  /// inherited from QObject and overridden here.
  /// @param _event the Qt Event structure
  /// \note these are part of the Qt API so can't be changed to the coding standard
  /// so it can't be called MousePressEvent )

  void mousePressEvent (
                        QMouseEvent *_event
                       );

  // \brief this method is called everytime the mouse button is released
  /// inherited from QObject and overridden here.
  /// @param _event the Qt Event structure
  /// \note these are part of the Qt API so can't be changed to the coding standard
  /// so it can't be called MousePressEvent )
  void mouseReleaseEvent (
                          QMouseEvent *_event
                         );

  /// @brief the timer event triggered from the timers
  /// @param _even the event of the timer triggered by Qt
  void timerEvent(
                  QTimerEvent *_event
                 );


};

#endif
