#include "Triangle.h"
#include <ngl/Util.h>
#include <ngl/VBOPrimitives.h>

Triangle::Triangle(
                    ngl::Vector _p0,
                    ngl::Vector _p1,
                    ngl::Vector _p2
                  )
{
  // set the default values
  m_u=0.0;
  m_v=0.0;
  m_w=0.0;
  m_v0=_p0;
  m_v1=_p1;
  m_v2=_p2;
  // calculate the edge
  m_edge1 = m_v1 - m_v0;
  m_edge2 = m_v2 - m_v0;
  // the center of the tri is the 3 verts average
  m_center=(m_v0+m_v1+m_v2)/3.0;
	m_hit=false;
	// create the m_points array for drawing the triangel
	m_points.push_back(m_v0.openGL()[0]);
	m_points.push_back(m_v0.openGL()[1]);
	m_points.push_back(m_v0.openGL()[2]);
	m_points.push_back(m_v1.openGL()[0]);
	m_points.push_back(m_v1.openGL()[1]);
	m_points.push_back(m_v1.openGL()[2]);
	m_points.push_back(m_v2.openGL()[0]);
	m_points.push_back(m_v2.openGL()[1]);
	m_points.push_back(m_v2.openGL()[2]);
	// now calculate the normal and create a normal array
	ngl::Vector normal=ngl::calcNormal(m_v0,m_v1,m_v2);

	m_normals.push_back(normal.m_x);
	m_normals.push_back(normal.m_y);
	m_normals.push_back(normal.m_z);
	m_normals.push_back(normal.m_x);
	m_normals.push_back(normal.m_y);
	m_normals.push_back(normal.m_z);
	m_normals.push_back(normal.m_x);
	m_normals.push_back(normal.m_y);
	m_normals.push_back(normal.m_z);
}

void Triangle::draw(
										const std::string &_shaderName,
										ngl::TransformStack &_transformStack
									 )
{

	// draw wireframe if hit
	if(m_hit)
	{
		glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
	}
	else
	{
	  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
	}


  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  // grab an instance of the primitives for drawing
  ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();

	// first we draw the triangle
	// we build up a vertex array for the lines of the start and end points and draw
	_transformStack.pushTransform();
	{
		shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",_transformStack.getCurrAndGlobal().getMatrix());

		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);



		glVertexPointer(3, GL_FLOAT, 0, &m_points[0]);
		glNormalPointer( GL_FLOAT, 0, &m_normals[0]);

		glDrawArraysInstancedARB(GL_TRIANGLES, 0, 3,1);
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
	}
	_transformStack.popTransform();

	// draw the cube to indicate vertex 0
	glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
	_transformStack.pushTransform();
	{
		_transformStack.getCurrentTransform().setPosition(m_v0);
		_transformStack.getCurrentTransform().setScale(0.06,0.06,0.06);
		shader->setShaderParamFromMatrix(_shaderName,"ModelMatrix",_transformStack.getCurrAndGlobal().getMatrix());
    prim->draw("cube");
	} // and before a pop
  _transformStack.popTransform();
	 // draw the hit point
	 if(m_hit)
   {
		 _transformStack.pushTransform();
		 {
			_transformStack.getCurrentTransform().setPosition(m_hitPoint);
			_transformStack.getCurrentTransform().setScale(2.0,2.0,2.0);
			shader->setShaderParamFromMatrix(_shaderName,"ModelMatrix",_transformStack.getCurrAndGlobal().getMatrix());
			prim->draw("smallSphere");
		 }
		_transformStack.popTransform();
   }

}


void Triangle::rayTriangleIntersect(
                                    ngl::Vector _rayStart,
                                    ngl::Vector _rayEnd
                                   )
{
  m_hit=false;
  // Calculate the ray direction
  ngl::Vector dir=_rayEnd-_rayStart;

  ngl::Vector tvec, pvec, qvec;
  float det, inv_det;
  // get the vector of the first edge
  pvec = dir.cross(m_edge2);
  // calculate the determinant
  det = m_edge1.dot(pvec);
  // if this is 0 no hit
  if (det > -0.00001f && det < 0.00001)
  {
    return ;
  }
  // get the inverse det
  inv_det = 1.0f / det;
  // calculate the 2nd vector
  tvec = _rayStart - m_v0;
  // get the dot product of this and inv det
  m_u = tvec.dot(pvec) * inv_det;
  // if out of range no hit
  if (m_u < -0.001f || m_u > 1.001f)
  {
    return;
  }
  // check the 2nd vector edge
  qvec = tvec.cross(m_edge1);
  // get the dot product
  m_v = dir.dot(qvec) * inv_det;
  // if out of range no hit
  if (m_v < -0.001f || m_u + m_v > 1.001f)
  {
    return;
  }
  // check the final value
  m_w = m_edge2.dot(qvec) * inv_det;
  // if less than 0 no hit
  if (m_w <= 0)
  {
    return;
  }
  // otherwise we are inside the triangle
  // so get the hit point
  // see http://softsurfer.com/Archive/algorithm_0105/algorithm_0105.htm#intersect_RayTriangle()
  // get intersect point of ray with triangle plane
  // calculate the normal
  ngl::Vector n=ngl::calcNormal(m_v0,m_v1,m_v2);
  float a = -n.dot(tvec);
  float b = n.dot(dir);
  float r=a/b;
  // intersect point of ray and plane
  m_hitPoint=_rayStart + r * dir;
  m_hit=true;
}











