#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Light.h"
#include "ngl/Matrix.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/Material.h"
#include "ngl/NGLInit.h"
#include "ngl/VBOPrimitives.h"
#include "ngl/ShaderManager.h"
#include "ngl/Random.h"
#include <boost/foreach.hpp>
#include "ngl/Util.h"


//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
									 int _numSpheres,
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

  // set this widget to have the initial keyboard focus
  setFocus();
	m_numSpheres=_numSpheres;
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // Now set the initial GLWindow attributes to default values
  // Roate is false
  m_rotate=false;
  // mouse rotation values set to 0
  m_spinXFace=0;
  m_spinYFace=0;
  m_origX=0;
  m_origY=0;
  m_sphereUpdateTimer=startTimer(30);
  m_animate=true;
	// now create the actual spheres for our program

	m_plane = new Plane(ngl::Vector(0,0,0),5,5);
	ngl::Vector pos;
	// now create the actual spheres for our program
	ngl::Random *rng=ngl::Random::instance();

	for(int i=0; i<_numSpheres; i++)
	{
		// use the uniform function from the rng class to get a nice
		// even random number
	  pos.set(rng->randomNumber(6),8,rng->randomNumber(6));
		// add the spheres to the end of the particle list
	  m_sphereArray.push_back(Sphere(pos,ngl::Vector(0,-1,0,0),0.2));
	}



}

GLWindow::~GLWindow()
{
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  std::cout<<"Shutting down NGL, removing VBO's and Shaders\n";
  Init->NGLQuit();
}

// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{
  // create the lights

  glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
  // enable depth testing for drawing
   glEnable(GL_DEPTH_TEST);
   // Now we will create a basic Camera from the graphics library
   // This is a static camera so it only needs to be set once
   // First create Values for the camera position
   ngl::Vector From(0,0,15);
   ngl::Vector To(0,0,0);
   ngl::Vector Up(0,1,0);
   ngl::NGLInit *Init = ngl::NGLInit::instance();
   #ifdef WIN32
   	glewInit();
   #endif
   Init->initGlew();
   m_cam= new ngl::Camera(From,To,Up,ngl::PERSPECTIVE);
   // set the shape using FOV 45 Aspect Ratio based on Width and Height
   // The final two are near and far clipping planes of 0.5 and 10
   m_cam->setShape(45,(float)720/576,0.05,350,ngl::PERSPECTIVE);
   // now to load the shader and set the values
   // grab an instance of shader manager
   ngl::ShaderManager *shader=ngl::ShaderManager::instance();
   // load a frag and vert shaders
   shader->loadShader("Blinn","shaders/Vertex.vs","shaders/Fragment.fs");
   // set this as the active shader
	 (*shader)["Blinn"]->use();
   // now pass the modelView and projection values to the shader
   shader->setShaderParamFromMatrix("Blinn","ViewMatrix",m_cam->getModelView());
   shader->setShaderParamFromMatrix("Blinn","projectionMatrix",m_cam->getProjection());
   shader->setShaderParam1i("Blinn","Normalize",true);
   shader->setShaderParam1i("Blinn","numLightsEnabled",1);
   shader->setShaderParam4f("Blinn","eye",0,0,15,1);

   glEnable(GL_DEPTH_TEST); // for removal of hidden surfaces

   ngl::VBOPrimitives *prim =  ngl::VBOPrimitives::instance();
   prim->createVBOSphere("sphere",1.0,40);


}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  glViewport(0,0,_w,_h);
  m_cam->setShape(45,(float)_w/_h,0.05,500,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	(*shader)["Blinn"]->use();
  shader->setShaderParamFromMatrix("Blinn","projectionMatrix",m_cam->getProjection());
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 // build our transform stack
  ngl::Transformation trans;
  // set the mouse rotation
  trans.setRotation(m_spinXFace,m_spinYFace,0);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);

  // grab an instance of the primitives for drawing
  ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();
  // Rotation based on the mouse position for our global
  // transform
  // grab an instance of the shader manager
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	(*shader)["Blinn"]->use();
	ngl::Material m(ngl::GOLD);
	m.use();

	//m_transformStack.loadGlobalAndCurrentMatrixToShader("Blinn","ModelMatrix");
  //shader->setShaderParamFromMatrix("Blinn","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());

	m_plane->draw("Blinn",m_transformStack);
	BOOST_FOREACH(Sphere s,m_sphereArray)
	{
		s.draw("Blinn",m_transformStack);
	}
}




//----------------------------------------------------------------------------------------------------------------------
void GLWindow::update()
{
	static int updateCount=0;
	BOOST_FOREACH(Sphere &s,m_sphereArray)
	{
		s.move();

	}
	spherePlaneCollide();

	if(++updateCount==20)
	{
		updateCount=0;
		ngl::Vector pos;
		BOOST_FOREACH(Sphere &s,m_sphereArray)
		{
			ngl::Random *rng=ngl::Random::instance();
			pos.set(rng->randomNumber(6),8,rng->randomNumber(6));
			s.set(pos,ngl::Vector(0,-1,0),0.2);
		}

	}

}



//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}

void GLWindow::timerEvent(
                          QTimerEvent *_event
                         )
{
	if(_event->timerId() == m_sphereUpdateTimer)
	{
		if (m_animate !=true)
		{
			return;
		}
		update();
		updateGL();
	}
}

void GLWindow::keyPress(QKeyEvent *_event)
{

  switch (_event->key())
  {
		case Qt::Key_Up    : m_plane->tilt(1.0,1,0); break;
		case Qt::Key_Down  : m_plane->tilt(-1.0,1,0); break;
		case Qt::Key_Left  : m_plane->tilt(1.0,0,1); break;
		case Qt::Key_Right : m_plane->tilt(-1.0,0,1); break;

	}

}


void GLWindow::spherePlaneCollide()
{
ngl::Vector p;
GLfloat D;
BOOST_FOREACH(Sphere &s , m_sphereArray)
{
  p=s.getNextPos();

  //If a collision is found we change the m_dir of the Sphere then Break
  //with the new point P
  D=m_plane->getNormal().dot(p);
  //Now Add the radius of the sphere to the offsett
  D+=s.getRadius();
  // If this is greater or equal to the BBox extent /2 then there is a collision
  //So we calculate the Spheres new direction
  if(D <=0.0)
  {

  //we on the plane now see if we hit it or not
    if(s.getPos().m_x > m_plane->getCenter().m_x-(m_plane->getWidth()/2.0) &&
       s.getPos().m_x < m_plane->getCenter().m_x+(m_plane->getWidth()/2.0) &&
       s.getPos().m_z > m_plane->getCenter().m_z-(m_plane->getDepth()/2.0) &&
       s.getPos().m_z < m_plane->getCenter().m_z+(m_plane->getDepth()/2.0)
      )
      {

        s.setDirection(m_plane->getNormal());
        s.setHit();
      }//end of hit test
    }
  }//end of each face test
}

