#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Light.h"
#include "ngl/Matrix.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/Material.h"
#include "ngl/NGLInit.h"
#include "ngl/VBOPrimitives.h"
#include "ngl/ShaderManager.h"
#include "ngl/Random.h"
#include <boost/foreach.hpp>
#include "ngl/Util.h"


//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

  // set this widget to have the initial keyboard focus
  setFocus();
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // Now set the initial GLWindow attributes to default values
  // Roate is false
  m_rotate=false;
  // mouse rotation values set to 0
  m_spinXFace=0;
  m_spinYFace=0;
  m_origX=0;
  m_origY=0;
  m_sphereUpdateTimer=startTimer(40);
  m_animate=true;
	// now create the actual spheres for our program

}

GLWindow::~GLWindow()
{
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  std::cout<<"Shutting down NGL, removing VBO's and Shaders\n";
  Init->NGLQuit();
}

// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{
  // create the lights

  glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
  // enable depth testing for drawing
   glEnable(GL_DEPTH_TEST);
   // Now we will create a basic Camera from the graphics library
   // This is a static camera so it only needs to be set once
   // First create Values for the camera position
   ngl::Vector From(0,0,-20);
   ngl::Vector To(0,0,0);
   ngl::Vector Up(0,1,0);
   ngl::NGLInit *Init = ngl::NGLInit::instance();
   #ifdef WIN32
   	glewInit();
   #endif
   Init->initGlew();
   m_cam= new ngl::Camera(From,To,Up,ngl::PERSPECTIVE);
   // set the shape using FOV 45 Aspect Ratio based on Width and Height
   // The final two are near and far clipping planes of 0.5 and 10
   m_cam->setShape(45,(float)720.0/576.0,0.5,150,ngl::PERSPECTIVE);
   // now to load the shader and set the values
   // grab an instance of shader manager
   ngl::ShaderManager *shader=ngl::ShaderManager::instance();
   // load a frag and vert shaders
   shader->loadShader("gl3xTest","shaders/Vertex.vs","shaders/Fragment.fs");
   // set this as the active shader
   shader->useShader("gl3xTest");
   // now pass the modelView and projection values to the shader
   shader->setShaderParamFromMatrix("gl3xTest","ViewMatrix",m_cam->getModelView());
   shader->setShaderParamFromMatrix("gl3xTest","projectionMatrix",m_cam->getProjection());
   shader->setShaderParam1i("gl3xTest","Normalize",true);
   shader->setShaderParam1i("gl3xTest","numLightsEnabled",1);

   glEnable(GL_DEPTH_TEST); // for removal of hidden surfaces

   ngl::VBOPrimitives *prim =  ngl::VBOPrimitives::instance();
   prim->createVBOSphere("sphere",1.0,40);
	 // create vectors for the position and direction
	 ngl::Vector pos;
	 ngl::Vector dir;
	 m_sphereArray.resize(4);
	 // now set two bigger spheres with no movement
	 pos.set(-10,0,0);
	 dir.set(0,0,0,0);
	 m_sphereArray[0].set(pos,dir,2.0);
	 pos.set(10,0,0);
	 dir.set(0,0,0,0);
	 m_sphereArray[1].set(pos,dir,2.0);
	 // and two smaller ones to move and bounce.
	 pos.set(-7,0,0);
	 dir.set(1,0,0,0);
	 m_sphereArray[2].set(pos,dir,1.0);
	 pos.set(7,0,0);
	 dir.set(-1,0,0,0);
	 m_sphereArray[3].set(pos,dir,1.0);


}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  glViewport(0,0,_w,_h);
  m_cam->setShape(45,(float)_w/_h,0.5,150,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  shader->setShaderParamFromMatrix("gl3xTest","projectionMatrix",m_cam->getProjection());
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 // build our transform stack
  ngl::Transformation trans;
  // set the mouse rotation
  trans.setRotation(m_spinXFace,m_spinYFace,0);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);

  // grab an instance of the primitives for drawing
  ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();
  // Rotation based on the mouse position for our global
  // transform
  // grab an instance of the shader manager
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	shader->use("gl3xTest");
	ngl::Material m(ngl::GOLD);
	m.use();

	m_transformStack.loadGlobalAndCurrentMatrixToShader("gl3xTest","ModelMatrix");
  shader->setShaderParamFromMatrix("gl3xTest","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
	BOOST_FOREACH(Sphere s, m_sphereArray)
	{
		s.draw("gl3xTest",m_transformStack);
	}

}




//----------------------------------------------------------------------------------------------------------------------
void GLWindow::update()
{
	m_sphereArray[2].move();
	m_sphereArray[3].move();
	checkCollisions();
}



//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}

void GLWindow::timerEvent(
                          QTimerEvent *_event
                         )
{
	if(_event->timerId() == m_sphereUpdateTimer)
	{
		if (m_animate !=true)
		{
			return;
		}
		update();
		updateGL();
	}
}

void GLWindow::keyPress(QKeyEvent *_event)
{

  switch (_event->key())
  {

	}

}


bool GLWindow::sphereSphereCollision(
																		 ngl::Vector _pos1,
																		 GLfloat _radius1,
																		 ngl::Vector _pos2,
																		 GLfloat _radius2
																		)
{
  // the relative position of the spheres
  ngl::Vector relPos;
  //min an max distances of the spheres
  GLfloat dist;
  GLfloat minDist;
  GLfloat len;
  relPos =_pos1-_pos2;
  // and the distance
  len=relPos.length();
  dist=len*len;
  minDist =_radius1+_radius2;
  // if it is a hit
  if(dist <=(minDist * minDist))
  {
    return true;
  }
  else
  {
    return false;
  }
}

void  GLWindow::checkCollisions()
{
//first check the small spheres against each other
bool collide =sphereSphereCollision(m_sphereArray[2].getPos(),m_sphereArray[2].getRadius(),
                  m_sphereArray[3].getPos(),m_sphereArray[3].getRadius());
if(collide== true)
	{
  m_sphereArray[2].reverse();
  m_sphereArray[3].reverse();
	}
// now for the little spheres against the big
collide =sphereSphereCollision(m_sphereArray[0].getPos(),m_sphereArray[0].getRadius(),
                  m_sphereArray[2].getPos(),m_sphereArray[2].getRadius());
// if we have a collision reverse the direction
if(collide== true)
{
  m_sphereArray[2].reverse();
}
collide =sphereSphereCollision(m_sphereArray[1].getPos(),m_sphereArray[1].getRadius(),
                  m_sphereArray[3].getPos(),m_sphereArray[3].getRadius());
if(collide== true)
{
  m_sphereArray[3].reverse();
}
}



